import { ShaderError} from "./gl_util";
import { Camera } from "./camera";
import * as G from "./glWrapper";
import { Drawable, RenderState } from "./drawable";
import { Model } from "./model";
import { mat4, vec3 } from "gl-matrix";
import {DirectionalLight, PhongProgram} from "./phongProgram";
import { UniformSet } from "./menu";


export class Drawer3D implements Drawable{
    gl : WebGL2RenderingContext;
    program: PhongProgram;
    
    dirLight: DirectionalLight;

    camera : Camera;
    teapot: Model | undefined;
    cube: Model | undefined;
    texture: G.Texture;
    constructor(gl: WebGL2RenderingContext){
        this.gl = gl;
        this.camera = new Camera([-20,1,-10]);
        this.camera.far = 1000;
        this.camera.near = 0.1;
        this.camera.fovY = Math.PI * 90 / 180;
        this.camera.aspect = 1;
        this.teapot = undefined;
        this.cube = undefined;
        this.texture = new G.Texture(gl);
        try{
            this.program = new PhongProgram(gl);
            const attr = this.program.getActiveAttributes(gl);
            console.log(attr);
        }
        catch(e){
            if(e instanceof ShaderError){
                console.log(e.info);
            }
            throw e;
        }
        this.dirLight = {
            ambientIntensity: 0.1,
            color: [1.0,1.0,1.0],
            diffuseIntensity: 0.5,
            direction: [0,1,0]
        };
    }
    async init(){
        const gl = this.gl;
        const teapot_url = new URL("../assets/models/teapot/teapot.obj",import.meta.url);
        this.teapot = await Model.loadFromOBJ(gl,teapot_url.href);
        const cube_url = new URL("../assets/models/cube/cube.obj",import.meta.url);
        this.cube = await Model.loadFromOBJ(gl,cube_url.href);
        console.log("loading model complete");
        this.teapot.ready(gl,this.program);
        this.cube.ready(gl,this.program);

        const texture_url = new URL("../assets/uv-grid.png",import.meta.url);
        const texture_image = G.makeImageElement(texture_url.href);
        texture_image.onload = ()=>{
            this.texture.setImage(gl,texture_image);
        }
        this.program.updateDirectionalLight(this.dirLight);
    }
    draw(gl: WebGL2RenderingContext,state:RenderState): void {
        this.camera.aspect = gl.canvas.width/gl.canvas.height;
        this.camera.UpdateProjectionMat();
        this.program.use(gl);
        this.program.viewMat = this.camera.viewMatrix;
        this.program.projectionMat = this.camera.projectionMatrix;
        this.program.setUniform3fv(gl,"cameraEyePosition",this.camera.pos);
        const colorTextureID = 0;
        this.texture.bind(gl,colorTextureID);
        this.program.colorTextureID = colorTextureID;
        if(this.teapot !== undefined){
            this.program.updateMaterial({
                shininess:20,
                specularIntensity:1.0
            });
            for(const pos of [[0,0,0],[0,0,25],[40,0,25]]){
                this.program.modelMat = mat4.fromTranslation(mat4.create(), vec3.fromValues(pos[0],pos[1],pos[2]));
                this.teapot.draw(gl);
            }
        }
        if(this.cube !== undefined){
            this.program.updateMaterial({
                shininess:20,
                specularIntensity:0.5
            });
            const modelMat = mat4.create();
            mat4.identity(modelMat);
            mat4.translate(modelMat,modelMat,[45,0,0]);
            mat4.scale(modelMat,modelMat,[5,5,5]);
            //mat4.rotateX(modelMat,modelMat,Math.PI * 30 / 180);
            this.program.modelMat = modelMat;
            this.cube.draw(gl);
        }
    }
    updateUniform(u:UniformSet){
        this.dirLight = {
            ambientIntensity: u.light_ambientIntensity,
            diffuseIntensity: u.light_diffuseIntensity,
            color: u.light_color,
            direction: u.light_direction
        };
        this.program.updateDirectionalLight(this.dirLight);
    }
}